/*---------------------------------------------------------------------------

  This program reads ASCII text from one or more files or from standard
  input and wraps it in PostScript code, with one-up, two-up, four-up, wide-
  print and long-print options.  With the appropriate compile-time flags the
  program may pipe the output either directly to a PostScript printer via
  lpr (Unix systems only) or to standard output.  [A runtime flag may allow
  the output to be directed into one or more files...someday.]

     Version:  Unix/VMS/OS2/MS-DOS; PostScript       Greg Roelofs
     Source:   gpr.c                                 5 March 1991

  To compile (old Unix):	cc -O -o gpr gpr.c
  To compile (GNU C):		gcc -O3 -Wall -o gpr gpr.c
  To compile (MinGW cross):	mingw32-gcc -O3 -Wall -o gpr.exe gpr.c

  Compile-time options (all but DEBUG are very old and/or deprecated):
     DEBUG			enable -D and -DD options for debugging
     DEBUG_SUBPAGE_ORIGINS	print calculated subpage origins for all modes
     OLDCC			include explicit declaration for getcwd();
				may need external strftime() as well
     OLD_SPRINTF		sprintf() returns char *, not int
     NEED_TAR_H			need tar.h header for getcwd() prototype
     SQUISH_NP			squish too-fat Sun NeWSprint 2.0 [deprecated]
     HORRIBLE_PMC_KLUDGE	add special-purpose "mode 11" [deprecated!!]

  Not yet implemented (possibly never to be implemented...):
     -o file    write to file (by default, output goes to stdout or printer)
     -T title   label to be used in header line in place of filename, if
                  input is from stdin
     proper FORTRAN print-manipulation options
     other stuff [see gpr.notes for random thoughts and changelog]

  ---------------------------------------------------------------------------*/


#define VERSION		"2.93"
#define VERSION_DATE	"6 Feb"
#define VERSION_YEAR	"2010"
#define PIPE

#include <stdio.h>		/* fopen(), getc(), fprintf(), fclose() */
#include <string.h>		/* strcpy() */
#include <time.h>		/* time(), localtime(), strftime() */
#include <ctype.h>		/* isdigit() */
#include <assert.h>		/* assert() */
#ifdef VMS
#  include <types.h>		/* stat():  dev_t */
#  include <stat.h>		/* stat() */
#  define ST_MTIME st_ctime
#  define MODERN
#  ifdef PIPE
#    undef PIPE			/* no popen() call or lpr daemon in VMS */
#  endif
#  define QUOT   	'\"'	/* for help screen */
#  define PATHSEP	']'	/* for testing */
#  include <unixlib.h>		/* getcwd() */
#  include <signal.h>		/* signal() */
#else /* Unix, OS/2 or MS-DOS */
#  include <sys/types.h>	/* stat():  dev_t */
#  include <sys/stat.h>		/* stat() */
#  define ST_MTIME st_mtime
#  if defined(MSDOS) || defined(__MSDOS__) || defined(__OS2__)
#    define MODERN
#    ifdef PIPE
#      undef PIPE		/* no popen() call or lpr daemon in MS-DOS */
#    endif
#    define QUOT   	' '	/* for help screen */
#    define PATHSEP	'\\'
#    include <stdlib.h>
#  else /* Unix */
#    define QUOT   	' '	/* for help screen */
#    define PATHSEP	'/'
#    ifndef OLDCC
#      ifdef NEED_TAR_H
#        include <tar.h>	/* getcwd() on some nasty system, apparently */
#      endif
#      include <unistd.h>	/* getcwd() */
#      include <stdlib.h>	/* getenv() */
#    else /* OLDCC */
       char *getcwd();
       char *getenv();
#    endif /* ?OLDCC */
#  endif
#endif /* ?VMS */

#if defined(__STDC__) || defined(__sgi)
#  define MODERN
#endif

#if defined(ultrix) || (defined(sun) && !defined(__SVR4))
#  define OLD_SPRINTF
#endif

#ifdef DEBUG
#  define DebugPrint(X)		if (debug) { fprintf X ; }
#else
#  define DebugPrint(X)
#endif

#define DEBUG_WARNING_OVERRUN	\
	"gpr warning:  truncating output to avoid buffer overrun.\n"
#define DEBUG_WARNING_UNDERRUN	\
	"gpr warning:  too many backspaces in input: would underrun buffer.\n"

#ifdef PIPE
#  ifdef OLD_EXAMPLES_OF_OTHER_PIPE_COMMANDS
#    define PIPE_LP_CMD	"lpr -Pps"	/* change as necessary */
#    define PIPE_LP_CMD	"rsh borgil -l aposky psp"
#    define PIPE_LP_CMD	"rsh mhd /usr/local/ipr -Lultrascript"
#    define PIPE_LP_CMD	"lpr -P%s"
#  endif
#  define PIPE_LP_CMD		"lpr -h -P%s"
#  define PIPE_USAGE		\
	"gpr [-#] [-bdHhnwz] [-g[#]] -Pprintername file [file ...]\n   or:  "
#  define PIPE_EXAMPLE		\
	"gpr -9 -Pps README foo.c foo.h bar.c bar.h Makefile"
#else
#  define PIPE_USAGE		""
#  define PIPE_EXAMPLE		""
#endif

#ifndef TRUE
#  define TRUE		1
#  define FALSE		0
#endif

#ifndef SEEK_SET
#  define SEEK_SET	0
#  define SEEK_CUR	1
#  define SEEK_END	2
#endif

#define NUL		0
#define SOH		1
#define BS		8
#define TAB		9
#define LF		10
#define FF		12
#define CR		13
#define DOS_EOF		26
#define VMS_RI		141

#define STREAMLF	0
#define RESTORE		1

#define BUFSIZE		2048	/* size of full pathname (header) and outbuf */
#define NUMBERS_WIDTH	7	/* prepended numbers:  5 digits + 2 spaces */

#define DATE_FMT	"%a %d-%b-%Y %H:%M"
#define DATE_LEN	21	/* length of result from DATE_FMT string */
#define FONT_STR	"/Courier findfont 12 scalefont setfont\n"
#define FONT_CPI	10	/* characters per (scaled) inch */
#define PS_PPI		72	/* points (PostScript units) per inch */

#define ODD_PAGE_MARGIN	14	/* or 15? (doubled == 29) */
#define EVEN_PAGE_DUPLEX_OFFSET	29	/* unscaled PS points */

typedef short   coords[2];

enum _orientation {
    PORTRAIT = 0,
    LANDSCAPE = 1
};

#define NUM_PAPER   5

struct paper_properties {
    char *name;
    int len_short_side;    /* PS coords (i.e., 1/72 of inch) */
    int len_long_side;     /* ditto */
    int gpr_pref_orient;   /* preferred orientation for layout of subpages */
} Paper[NUM_PAPER] = {

/*     name     short  long  pref. orient.  */
/*   ---------  -----  ----  -------------  */
    {"8.5x11",   612,   792,   PORTRAIT     },   /* US letter */
    {"8.5x14",   612,  1008,   PORTRAIT     },   /* US legal */
    { "11x17",   792,  1224,   LANDSCAPE    },   /* US ledger/tabloid */
    {"210x297",  595,   842,   PORTRAIT     },   /* ISO A4: ratio = 1:sqrt(2) */
    {"297x420",  842,  1191,   LANDSCAPE    }    /* ISO A3: ratio = 1:sqrt(2) */
};                       /* see http://www.cl.cam.ac.uk/~mgk25/iso-paper.html */

// subpage bounding box with header (SCALED):
//    (0,0) -> (CharMax*PS_PPI/FONT_CPI + 12, YINC*LineMax + 36)
//    [ACTUAL:  multiply by scaling factor]

// subpage bounding box WITHOUT header (SCALED):
//    (0,0) -> (CharMax*PS_PPI/FONT_CPI + 12, YINC*LineMax + 8)
//    [ACTUAL:  multiply by scaling factor]

// horizontal, vertical gap between logical subpages (SCALED):
//    24

struct layout {
    int scale;      /* PostScript scaling factor * 10000 (i.e., fixed-point) */
    coords subpage_geom;   /* subpage layout--e.g., 3x1 */
    short row_major;       /* consecutive subpages go left-to-right? */
    short char_max;
    short line_max;
    short rotate;
    coords phy_orig;	/* lower left corner of BOX, not upper left of text! */
    coords phy_orig2;	/* currently only for landscape modes when duplexing */
} Layout[] = {

/*           +-- for single-size paper; 2*horiz for double-size              */
/*           |                       +-- could use PORTRAIT/LANDSCAPE here   */
/*         subpg  subpg    W x H     |                                       */
/*   scale  geom RowMaj  chrs lns   rot.  phy_orig  phy_orig2                */
/*   ----- ----- ------  ---- ---  -----  --------  ---------                */
    {4545, {1,1}, TRUE,  165, 136, FALSE,  {47,17},    {0,0}},  /*  0 */
    {9091, {1,1}, TRUE,   80,  66, FALSE,  {52,17},    {0,0}},  /*  1 */
    {6066, {2,1}, TRUE,   80,  66, TRUE,  {571,47}, {37,803}},  /*  2 */
    {6066, {1,1}, FALSE, 165,  66, TRUE,  {571,47}, {37,803}},  /*  3 */
    {4545, {2,2}, TRUE,   80,  66, FALSE,  {47,12},    {0,0}},  /*  4 */
    {5675, {1,2}, FALSE, 132,  51, FALSE,  {47,17},    {0,0}},  /*  5 */
    {5675, {1,1}, TRUE,  132, 108, FALSE,  {47,17},    {0,0}},  /*  6 */
    {6800, {1,1}, FALSE, 147,  66, TRUE,  {586,47}, {25,801}},  /*  7 */
    {4545, {2,2}, FALSE,  80,  66, FALSE,  {47,12},    {0,0}},  /*  8 */
    {4545, {2,1}, TRUE,   80, 136, FALSE,  {47,17},    {0,0}},  /*  9 */
    {3030, {3,1}, TRUE,   80, 205, FALSE,  {47,17},    {0,0}}   /* 10 */
#ifdef HORRIBLE_PMC_KLUDGE
  , {5700, {4,1}, TRUE,   43,  83, TRUE,  {598,31}, {37,756}}   /* 11 */
#endif
};

/*  Note that phy_orig/PhysOrigin is used before rotation and scaling.  */
/*  ptswide = CharMax * PS_PPI / FONT_CPI;  */

#define Scale           Layout[mode].scale
#define SubpageGeom     Layout[mode].subpage_geom
#define RowMajor        Layout[mode].row_major
#define CharMax         Layout[mode].char_max
#define LineMax         Layout[mode].line_max
#define Rotate          Layout[mode].rotate
#define PhysOrigin      Layout[mode].phy_orig
#define PhysOrigin2     Layout[mode].phy_orig2

/* more global variables (yeah, yeah, ick...but hey--it's not like we're
 * trying to be reentrant) */

#define SUBPAGE_MAX  16		/* max number of subpages per side of paper */
#define SUBPAGE_GAP  24		/* (scaled) PostScript points between boxes */
#define YINC         12		/* line height in (scaled) PostScript points */



     /* [nominal box-height plus header for 66 rows == 828 == 12*(66 + 3)] */
     /* [nominal box-width for 80 columns == 588 == 80*72/10 + 12] */
     /* [612 == extra 24 scaled points == scaled third-inch ~= 3.3 chars] */
     /* 588 + 24 + 588 = 1200 is equivalent to box with 165 chars inside */

/* {x,y} scaled offsets between origins of subpages (assuming mult.) => lo*
{1224,1692}   horiz = (165*72/10 + 12) + 24    vert = 12*(136 + 3) + 24
{612,852}              ===                                ===
{612,852}
 */

char    fullname[BUFSIZE];		/* for header only */
char    filename[256], when[40];
int     mode=1, tabsize=8, border=TRUE, gray=1, header=TRUE, numbers=FALSE;
int     wrap=TRUE, zeropage=FALSE, doublesided=FALSE, SubpageMax;
int     SubpageGeomHoriz, subpage_bbox_width, subpage_bbox_height;
coords  SubpageOrigin[SUBPAGE_MAX];
time_t  CurTim;
FILE    *infile, *outfile;
#ifdef DEBUG
int     debug = 0;
#endif
#ifdef HORRIBLE_PMC_KLUDGE
int     pmc_hack = FALSE;
#endif
#ifdef SQUISH_NP
int     squish_np = FALSE;
#endif

/* prototypes */

void   print_file (void);
int    PS_strncpy (char *dest, const char *src, int n);
#ifdef VMS
void   handler (void);
int    vms_getc (FILE *fp);
int    change_file_attributes (char *fname, int action);
#endif





/******************/
/*  Main program  */
/******************/

int main(int argc, char *argv[])
{
    int           i, j, error=0, c, c1, n, help=FALSE, version=0, subpg;
    int           num_modes = sizeof(Layout) / sizeof(struct layout);
#ifdef PIPE
    int           pipelpr=FALSE;
    char          *printer="lp", printer_pipe[256];
#endif
    char          *p, *q, *r;
    char          pathname[1024];   /* for header only */
    struct stat   statbuf;
    FILE          *usagefp;


/*---------------------------------------------------------------------------
    Set up interrupt handler so file attributes get restored properly.
  ---------------------------------------------------------------------------*/

#ifdef VMS
    signal(SIGINT, handler);
#endif

/*---------------------------------------------------------------------------
    Parse command line for '-' options.
  ---------------------------------------------------------------------------*/

    if (argc == 1) {       /* just "gpr" */
        --argc;
        if (isatty(0) || ((c = getc(stdin)) == EOF))
            help = TRUE;   /* stdin == console or /dev/null */
        else
            ungetc(c, stdin);
    } else {
        while (--argc > 0  &&  (*++argv)[0] == '-') {
            while ((c = *++(argv[0]))) { /* argv[0] is no longer program name */
                switch (c) {
                    case 'b':		/* disable sub-page borders */
                        border = FALSE;
                        break;
                    case 'd':		/* adjust for double-sided output */
                        doublesided = TRUE;
                        break;
#ifdef DEBUG
                    case 'D':		/* debug mode, obviously! */
                        ++debug;
                        break;
#endif
                    case 'g':		/* set header grayness level (0-9) */
                        c1 = argv[0][1];   /* get next char w/o changing ptr */
                        if ((c1 >= '0') && (c1 <= '9')) {
                            gray = atoi(argv[0]+1);   /* ptr to digit string */
                            argv[0][1] = '\0';        /* overwrite digit */
                        } else
                            gray = 3;   /* default value */
                        break;
                    case 'h':		/* just print help message */
                        help = TRUE;
                        break;
                    case 'H':		/* suppress header lines */
                        header = FALSE;
                        break;
                    case 'n':		/* prepend line numbers */
                        numbers = TRUE;
                        break;
#if 0
                    case 'o':		/* write to file */
                        file_output = TRUE;
                        break;
#endif
#ifdef PIPE
                    case 'P':		/* pipe to printer */
                        pipelpr = TRUE;
                        if (argv[0][1]) {
                            printer = &argv[0][1];
                            while (*++(argv[0]))
                                ;
                            --(argv[0]);
                        } else if (argc > 1) {
                            --argc;
                            ++argv;
                            if (argv[0][0] != '-') {
                                printer = argv[0];
                                while (*++(argv[0]))
                                    ;
                                --(argv[0]);
                            }
                            /* else use default printer name, "lp" */
                        }
                        break;
#endif /* PIPE */
#ifdef SQUISH_NP
                    case 's':		/* squish too-fat Sun NeWSprint 2.0 */
                        squish_np = TRUE;
                        break;
#endif
                    case 't':		/* tab spacing */
                        c1 = argv[0][1];   /* get next char w/o changing ptr */
                        if ((c1 >= '1') && (c1 <= '9')) {
                            tabsize = atoi(argv[0]+1);   /* ptr to digits */
                            argv[0][1] = '\0';        /* overwrite 1st digit */
                        } else
                            tabsize = 8;              /* default value */
                        break;
                    case 'v':		/* print version info and exit */
                        ++version;
                        break;
                    case 'w':		/* turn off wrapping of long lines */
                        wrap = FALSE;
                        break;
                    case 'z':		/* print blank "page zero" */
                        zeropage = TRUE;
                        break;
#if 0
                    case '0':		/* one double-length, extra-wide page */
                    case '1':		/* one logical page */
                    case '2':		/* two logical pages */
                    case '3':		/* one extra-wide logical page */
                    case '4':		/* four logical pages, 1 next to 2 */
                    case '5':		/* two wide logical pages*/
                    case '6':		/* one wide, extra-long logical page */
                    case '7':		/* one wide logical page */
                    case '8':		/* four logical pages, 1 above 2 */
                    case '9':		/* two double-length logical pages */
                        mode = c - '0';
                        break;
                -Nlabel    label to be used in header line
#endif
                    default:
                        if ((c >= '0') && (c <= '9')) {
                            mode = atoi(argv[0]);   /* ptr to digits */
                            while (isdigit(argv[0][1]))
                                ++(argv[0]);
                            if (mode >= num_modes) {
                                fprintf(stderr,
                                  "error:  %d is an invalid mode\n", mode);
                                ++error;
                            }
                        } else {
                            fprintf(stderr,
                              "error:  `%c' is not a valid option\n", c);
                            ++error;
                        }
                        break;
                }  /* end switch over option types */
            }  /* end while-loop over options within single argument */
        }  /* end while-loop over option-arguments */
    }  /* end if (argc != 1) */

/*---------------------------------------------------------------------------
    Print usage screen, if any errors encountered.
  ---------------------------------------------------------------------------*/

    if (help && !error)
        usagefp = (FILE *)stdout;
    else
        usagefp = (FILE *)stderr;

    if (help || error) {
        fprintf(usagefp,
"gpr v%s of %s %s:  simple, multi-format ASCII-to-PostScript converter.\n"
"Copyright 1991-%s Greg Roelofs (newt@pobox.com).\n"
"\n"
"Usage:  %sgpr [-#] [-bdHhnwz] [-g[#]] file [file ...] > files.ps\n"
"   or:  ... | gpr [-#] [-bdHhnwz] [-g[#]] | ...\n"
"Example:  %s\n"
"     or:  egrep 'Minneapolis|St. Paul' phone.book | gpr -2g9 > twin-cities.ps\n"
"\n"
"Page layout modes (number of logical pages per physical page, columns x lines):\n"
"  -1  one 80 x 66 (default)              -6  one 132 x 108\n"
"  -2  two 80 x 66 [landscape]            -7  one 147 x 66 [landscape]\n"
"  -3  one 165 x 66 [landscape]           -8  four 80 x 66; 1 & 3 above 2 & 4\n"
"  -4  four 80 x 66; 1 & 2 above 3 & 4    -9  two 80 x 136; the `green' mode!\n"
"  -5  two 132 x 51                       -0  one 165 x 136\n"
"\n"
"  -10 three 80 x 204; the `super-green' mode!\n"
"\n"
"  -w  suppress wrapping of long lines; truncate   -n  prepend line numbers\n"
"  -b  suppress border boxes around sub-pages     %c-H%c suppress header lines\n"
"  -z  include blank \"page zero\" if multi-page     -h  print help (this message)\n"
"  -d  adjust even-page margins for double-sided (duplex) output\n"
"\n"
"  -g#  header color:  0 white, 1 boxed (default), 3 gray (-g default), 9 black\n"
"  -t#  tab spacing (default = 8)\n"
"\n"
"file:  name(s) of file(s) to be converted/printed; if missing, standard input\n"
"       is assumed.  Output goes to stdout; pipe to printer or redirect to file."
"\n",
          VERSION, VERSION_DATE, VERSION_YEAR, VERSION_YEAR,
          PIPE_USAGE, PIPE_EXAMPLE, QUOT, QUOT);

#ifdef VMS
        fprintf(usagefp, "\n"
"Use ^C instead of ^Y to kill the program, or else the file attributes will\n"
"not be restored properly.\n");
#endif
        exit(error);
    }

/*---------------------------------------------------------------------------
    Print version info and exit, if requested.
  ---------------------------------------------------------------------------*/

    /* GRR TODO/FIXME:  add compiler info here */
    if (version) {
        if (version > 1) {
            printf("%s\n", VERSION);
        } else {
            printf("gpr v%s of %s %s (no further info available)\n", VERSION,
              VERSION_DATE, VERSION_YEAR);
        }
        exit(0);
    }

/*---------------------------------------------------------------------------
    Calculate page geometry according to paper size, orientation, mode, etc.
  ---------------------------------------------------------------------------*/

#if 0

				GRR NOTES


 - basic corner offset == somewhere between 18 and 36
	[is this lower left text or corner of boundary box?  want latter]
 - want to be able to reproduce basic usage info in usage screen for each
    paper size:

US Letter:
/*
Page layout modes (number of logical pages per physical page, columns x lines):
  -1  one 80 x 66 (default)              -6  one 132 x 108
  -2  two 80 x 66 [landscape]            -7  one 147 x 66 [landscape]
  -3  one 165 x 66 [landscape]           -8  four 80 x 66; 1 & 3 above 2 & 4
  -4  four 80 x 66; 1 & 2 above 3 & 4    -9  two 80 x 136; the `green' mode!
  -5  two 132 x 51                       -0  one 165 x 136

  -10 three 204 x 80; the `super-green' mode!
 */


 - assume basic corner offset = (32,32) (?) (+ local adjustment per mode)
 - given paper size, printer orientation (P/L) == preferred orientation...

 - if orientation == preferred orientation, AND not landscape mode(!), => no
   long/short offset or rotation, just -d offset (+/- x)
 - treat 11x17, A3 (only two landscape formats) just as double-size pages?
   => new paper property:  "single" vs. "double" (correlated with landscape)
 - keep track of row-major vs. column major for modes 4,8 and double-size paper
 - allow 8(?) subpages MAX per side => single struct
   - keep gross geom. instead and calculate num subpages in each direction?

 - if orient. mismatch (and -d mode!), then:
   - choose long or short axis (+ rot. direction) based on odd/even page =>
     major offset
   - -d offset = +/- y if rolled into main translation, else subs. +/- x offset
     (test this!)
   - logical offset to top of subpage is _scaled_ (and depends on lines,
     subpages)
 - basic geom/layout per paper size:
   - scale (slight adjustments allowed)
   - chars x lines per subpage (and which one(s) is "soft")	\_ related
   - preferred subpages per column (soft?) and row (soft?)	/

 - [possibly] loop over paper sizes
   - loop over modes
     - for each mode, have:
       - scale (within ~ 2%)
       - subpage layout (possibly doubled in horizontal direction)
       - preferred lines, width
     - calculate box width
     ...








 - encode paper size in headers, as with acroread, etc. (PS level 3 required?)
 - code is now rewritten to use lower left corner as origin; test landscape
   mode(s), then calculate four corners of subpage bounding boxes => use for
   centering subpages on page


struct paper_properties {
    char *name;
    int len_short_side;    /* PS coords (i.e., 1/72 of inch) */
    int len_long_side;     /* ditto */
    int gpr_pref_orient;   /* preferred orientation for layout of subpages */
} Paper[NUM_PAPER] = {

/*     name     short  long  pref. orient.  */
/*   ---------  -----  ----  -------------  */
    {"8.5x11",   612,   792,   PORTRAIT     },   /* US letter */
    {"8.5x14",   612,  1008,   PORTRAIT     },   /* US legal */
    { "11x17",   792,  1224,   LANDSCAPE    },   /* US ledger/tabloid */
    {"210x297",  595,   842,   PORTRAIT     },   /* ISO A4: ratio = 1:sqrt(2) */
    {"297x420",  842,  1191,   LANDSCAPE    }    /* ISO A3: ratio = 1:sqrt(2) */
};                       /* see http://www.cl.cam.ac.uk/~mgk25/iso-paper.html */

#endif /* 0 */

#if 0
OK      subpage origins, mode  0:  {0,0} 
OK      subpage origins, mode  1:  {0,0} 
OK      subpage origins, mode  2:  {0,0} {612,0} 
OK      subpage origins, mode  3:  {0,0} 
OK      subpage origins, mode  4:  {0,852} {612,852} {0,0} {612,0} 
DIFFERS subpage origins, mode  5:  {0,648} {0,0} 
                                   {0,667} {0,0} originally
OK      subpage origins, mode  6:  {0,0} 
OK      subpage origins, mode  7:  {0,0} 
OK      subpage origins, mode  8:  {0,852} {0,0} {612,852} {612,0} 
OK      subpage origins, mode  9:  {0,0} {612,0} 
OK      subpage origins, mode 10:  {0,0} {612,0} {1224,0} 
DIFFERS subpage origins, mode 11:  {0,0} {345,0} {690,0} {1035,0} 
                                   {0,0} {336,0} {672,0} {1008,0} 
#endif // 0


    // calculate subpage origins:

#ifdef DEBUG_SUBPAGE_ORIGINS
    for (mode = 0;  mode < num_modes;  ++mode) {
#endif
    SubpageGeomHoriz = SubpageGeom[0];
    // if (paper.doublesize)
    //     SubpageGeomHoriz *= 2;
    SubpageMax = SubpageGeomHoriz * SubpageGeom[1];
    assert(SubpageMax <= SUBPAGE_MAX);

    DebugPrint((stderr, "subpage origins, mode %2d:  ", mode));

    // GRR FIXME:  want to adjust subpage_bbox_height if header == FALSE?
    //             (center vertically on physical page?)
    subpage_bbox_width = CharMax*PS_PPI/FONT_CPI + 12;
    subpage_bbox_height = LineMax*YINC + 36;
    subpg = 0;   /* cycles through logical sub-pages */

    if (RowMajor) {	// row-major:  inner loop == columns (same row) == x
        for (j = SubpageGeom[1] - 1;  j >= 0;  --j) {   // top-down, not bot-up
            for (i = 0;  i < SubpageGeomHoriz;  ++i, ++subpg) {
                SubpageOrigin[subpg][0] = i*(subpage_bbox_width + SUBPAGE_GAP);
                SubpageOrigin[subpg][1] = j*(subpage_bbox_height + SUBPAGE_GAP);
                DebugPrint((stderr, "{%d,%d} ", SubpageOrigin[subpg][0],
                  SubpageOrigin[subpg][1]));
            }
        }
    } else {		// column-major:  inner loop == rows (same column) == y
        for (i = 0;  i < SubpageGeomHoriz;  ++i) {
            for (j = SubpageGeom[1] - 1;  j >= 0;  --j, ++subpg) {  // top-down
                SubpageOrigin[subpg][0] = i*(subpage_bbox_width + SUBPAGE_GAP);
                SubpageOrigin[subpg][1] = j*(subpage_bbox_height + SUBPAGE_GAP);
                DebugPrint((stderr, "{%d,%d} ", SubpageOrigin[subpg][0],
                  SubpageOrigin[subpg][1]));
            }
        }
    }

    DebugPrint((stderr, "\n"));
    assert(subpg == SubpageMax);
#ifdef DEBUG_SUBPAGE_ORIGINS
    }
    exit(0);
#endif

/*---------------------------------------------------------------------------
    Adjust any mode-specific or mildly incompatible options.  (Crude...)
  ---------------------------------------------------------------------------*/

    if (zeropage && (SubpageMax == 1))
        zeropage = FALSE;

#ifdef HORRIBLE_PMC_KLUDGE	/* for a long-dead telephone list */
    if (mode == 11) {
        border = FALSE;
        header = FALSE;
        pmc_hack = TRUE;
    }
#endif

/*---------------------------------------------------------------------------
    Open output file as needed and get current working directory for pre-
    pending to filenames.  This probably gets pretty ugly under VMS (not
    tested).
  ---------------------------------------------------------------------------*/

#ifdef PIPE
    if (pipelpr) {
        sprintf(printer_pipe, PIPE_LP_CMD, printer);
        if ((outfile = popen(printer_pipe, "w")) == NULL) {
            fprintf(stderr, "gpr error:  cannot open pipe to printer\n");
            exit(1);
        }
    } else
#endif
    outfile = stdout;

    if (header) {
        if ((q = getcwd(pathname, 1024)) == NULL) {
            pathname[0] = '.';
            pathname[1] = '\0';
        } else {
            char *homedir, homepathname[1024];
            int len;

            if ((homedir = getenv("HOME")) != NULL  &&  chdir(homedir) == 0) {
                q = getcwd(homepathname, 1024);
                if (chdir(pathname) != 0) {
                    fprintf(stderr, "gpr warning:  oops, couldn't chdir() back "
                      "to %s for some reason (shouldn't matter, though)\n",
                      pathname);
                }
                if (q && strncmp(pathname, q, (len = strlen(q))) == 0) {
                    p = pathname + len;  /* typically "/foo/bar/..." */
                    q = pathname;
                    *q++ = '~';
                    while ((*q++ = *p++));  /* now "~/foo/bar/..." */
                }
                /* else no change */
            }
#ifdef __MINGW32__
/*
            || defined(_WIN32) || defined(WIN32) || defined(__WIN32__)

            -D__GNUC__=2 -D__GNUC_MINOR__=95 -Di386 -D_WIN32 -DWIN32
            -D__WIN32__ -D__MINGW32__ -D__MSVCRT__ -DWINNT -D_X86_=1
            -D__STDC__=1 -D__stdcall=__attribute__((__stdcall__))
            -D_stdcall=__attribute__((__stdcall__))
            -D__cdecl=__attribute__((__cdecl__))
            -D__declspec(x)=__attribute__((x)) -D__i386__
            -D_WIN32 -D__WIN32__ -D__WIN32__ -D__MINGW32__
            -D__MSVCRT__ -D__WINNT__ -D_X86_=1 -D__STDC__=1
            -D__stdcall=__attribute__((__stdcall__))
            -D___stdcall__=__attribute__((__stdcall__))
            -D__cdecl=__attribute__((__cdecl__))
            -D__declspec(x)=__attribute__((x)) -D__i386 -D__WIN32
            -D__WINNT -D___stdcall=__attribute__((__stdcall__))
            -D__OPTIMIZE__
 */
            /* MinGW version of getcwd() returns backslashes, not forward
             * slashes, so convert for consistency.  (See also PS_strncpy()
             * calls below, which cope with non-DIRSEP backslashes and
             * parens.) */
            q = pathname - 1;
            while (*++q) {
                if (*q == '\\')
                    *q = '/';
            }
#endif
        }
    }

/*---------------------------------------------------------------------------
    Any arguments remaining (i.e., not preceded by '-') must be the names of
    the ASCII files to be converted, so start doing that now.
  ---------------------------------------------------------------------------*/

    time(&CurTim);   /* for PS header, stdin "filetime" */

    if (argc > 0) {
        for (n = 0;  n < argc;  ++n) {
            q = argv[n];
            r = filename;
            while ((*r++ = *q++));	/* copy argument into filename[] */
            if (stat(filename, &statbuf)) {
                ++error;
                continue;   /* error, so just go on to next file */
            }
#ifdef VMS
            change_file_attributes(filename, STREAMLF);
#endif
            if ((infile = fopen(filename, "rb")) == NULL) {
                ++error;
#ifdef VMS
                change_file_attributes(filename, RESTORE);
#endif
                continue;   /* can't open, so try for next */
            }
            if (header
#ifdef HORRIBLE_PMC_KLUDGE
                       || pmc_hack
#endif
                                  ) {
                strftime(when, 40, DATE_FMT, localtime(&statbuf.ST_MTIME));
                p = fullname;
#ifndef VMS
                if (*argv[n] != PATHSEP) { /* not an absolute pathname */
#endif
                    q = pathname;
                    while ((*p++ = *q++)); /* copy path into fullname */
                    if (p[-2] == PATHSEP)  /* ended w/ separator:  prob. root */
                        --p;
#ifndef VMS
                    else
                        p[-1] = PATHSEP;   /* replace NULL w/ path separator */
                }
#endif
                q = argv[n];
                while ((*p++ = *q++));	   /* copy argument into fullname[] */
            }
            print_file();
            fclose(infile);
#ifdef VMS
            change_file_attributes(filename, RESTORE);
#endif
        }
    } else {	/* no command-line arguments left:  must be standard input */
        if (header) {
            strcpy(fullname, "standard input");
            strftime(when, 40, DATE_FMT, localtime(&CurTim));
        }
#ifdef HORRIBLE_PMC_KLUDGE
        if (pmc_hack) {
         /* strftime(when, 40, "%H:%M, %A, %d %B %Y", localtime(&CurTim)); */
            strftime(when, 40, "%A, %d %B %Y", localtime(&CurTim));
        }
#endif
        infile = stdin;
        print_file();
    }

#ifdef PIPE
    if (pipelpr)
        pclose(outfile);  /* does no harm if not popen'd (i.e., stdout) */
#endif

#ifdef VMS
    exit(0);              /* VMS interprets non-zero error codes incorrectly */
#else
    return error;
#endif

} /* end main() */





#ifdef SQUISH_NP  /* see definitions of show_string vs. ashow_string below */
#  define SQUISH	squish_np? ashow_string :
#else
#  define SQUISH
#endif

/***************************/
/*  Function print_file()  */
/***************************/

void print_file(void)
{
    unsigned char outbuf[BUFSIZE], *outptr;
    int num_offset = numbers? NUMBERS_WIDTH : 0;
    int ptswide = CharMax * PS_PPI / FONT_CPI;
    int subpagenum=0, lineno=0, phys_pagenum=0, wrapped=FALSE;
    char *show_string = "%d %d moveto (%s) show\n";
#ifdef SQUISH_NP
    char *ashow_string = "%d %d moveto -.18 0 (%s) ashow\n";
#endif
#ifdef VMS
    int (*get_c)();
/*  int fortran;  */  /* not actually used for anything */
#endif
#if 0
    int file_output=FALSE;
#endif


/*---------------------------------------------------------------------------
    Special VMS processing section to detect obnoxious semi-ASCII formats.
  ---------------------------------------------------------------------------*/

#ifdef VMS
/*  fortran = FALSE;  */
    get_c = fgetc;
    fread(outbuf, 1, 4, infile);	/* read first four bytes */
#if 0
    for (c = 0;  c < 4;  ++c)		/* strip high bits, if sign-extended */
        outbuf[c] &= 0xFF;
#endif
    DebugPrint((stderr, "FIRST FOUR BYTES:  %x %x %x %x\n", outbuf[0],
      outbuf[1], outbuf[2], outbuf[3]));

    if ((outbuf[0] == NUL) && (outbuf[1] == NUL) && (outbuf[3] == NUL)) {
/*      fortran = TRUE;  */
        get_c = vms_getc;
        fseek(infile, 2L, SEEK_SET);
        DebugPrint((stderr, "SPECIAL FILE:  FORTRAN\n"));
    } else if ((outbuf[2] == SOH) && (outbuf[3] == VMS_RI)) {
        get_c = vms_getc;
        rewind(infile);
        DebugPrint((stderr, "SPECIAL FILE:  PRINTFILE\n"));
    } else if (outbuf[1] == NUL) {
        get_c = vms_getc;
        rewind(infile);
        DebugPrint((stderr, "SPECIAL FILE:  VARIABLE-LENGTH RECORDS\n"));
    } else
        rewind(infile);
#endif /* VMS */

/*---------------------------------------------------------------------------
    Print PostScript header (including all-important signature bytes "%!").
  ---------------------------------------------------------------------------*/

    fprintf(outfile, "%%!PS-Adobe-1.0\n");
    fprintf(outfile, "%%%%DocumentFonts: Courier\n");
#ifdef HORRIBLE_PMC_KLUDGE
    if (pmc_hack)
        fprintf(outfile, "%%%%+ Times-Bold\n");
#endif
#ifdef DEBUG
    if (debug)
        fprintf(outfile, "%%%%+ Times-Bold\n");
#endif
    fprintf(outfile, "%%%%Pages: (atend)\n");
    fprintf(outfile, "%%%%Title: %s\n", filename);
    fprintf(outfile, "%%%%Creator: gpr v%s of %s %s\n", VERSION, VERSION_DATE,
      VERSION_YEAR);
    fprintf(outfile, "%%%%CreationDate: %s", ctime(&CurTim));
    fprintf(outfile, "%%%%EndComments\n");
    fprintf(outfile, "%%\n%%   program:   gpr\n");
    fprintf(outfile,
      "%%   function:  general-purpose ASCII-to-PostScript converter\n");
    fprintf(outfile, "%%   mnemonic:  \"Giant Panda Rollerskates\"  :-)\n");
    fprintf(outfile, "%%   author:    Greg Roelofs\n");
    fprintf(outfile, "%%   created:   5 March 1991\n%%\n");

    /* these may be useful for other things in the future (e.g.,
     * right-justifying the page number in a non-Courier header font),
     * so we'll toss them in for all modes, regardless of usage [20030304] */
    fprintf(outfile,
      "/rightshow { dup stringwidth pop neg 0 rmoveto show } def\n");
    fprintf(outfile,
      "/centershow { dup stringwidth pop 2 div neg 0 rmoveto show } def\n");

    fprintf(outfile, "save\n");   /* make sure we restore state completely */

/*---------------------------------------------------------------------------
    Loop indefinitely over physical pages (EOF exits routine).
  ---------------------------------------------------------------------------*/

    for (;;) {
        int subpage_counter = 0;   /* cycles through logical sub-pages */
        int xindent=0, yindent=0;  /* if duplex, shift even pages for binding */
        int end_physical_page = FALSE;

        if ((phys_pagenum == 0) && zeropage)
            subpage_counter = 1;
        ++phys_pagenum;

#ifdef DEBUG
        if (debug) {
            fprintf(outfile, "save\n");
            fprintf(outfile, "0.93 0.93 0.93 setrgbcolor\n");
            fprintf(outfile, "/Times-Bold findfont 120 scalefont setfont\n");
            fprintf(outfile, "306 684 moveto (mode %d) centershow\n", mode);
            fprintf(outfile, "restore\n");
        }
#endif

        /* if doubled-sided output and we're on an even page, indent */
        if (doublesided && !(phys_pagenum & 1)) {
            if (Rotate) {                           /* landscape orientation */
                yindent = -EVEN_PAGE_DUPLEX_OFFSET;
            } else {
                xindent = -EVEN_PAGE_DUPLEX_OFFSET;
            }
        }

        /* if doublesided and landscape and even page, translate/rotate in
         * opposite direction so tops of _logical_ pages (subpages) will be
         * at same edge of paper */
// GRR:  may allow slight difference in xscale, yscale here if A4, for example
        if (doublesided && !(phys_pagenum & 1) && Rotate) {
            fprintf(outfile,
              "save %d %d translate -90 rotate .%04d .%04d scale\n",
              PhysOrigin2[0] + xindent, PhysOrigin2[1] + yindent,
              Scale, Scale);
        } else {
            fprintf(outfile, "save %d %d translate%s .%04d .%04d scale\n",
              PhysOrigin[0] + xindent, PhysOrigin[1] + yindent,
              Rotate? " 90 rotate" : "", Scale, Scale);
            /* (e.g., "save 567 27 translate 90 rotate .6066 .6066 scale\n") */
        }

        fprintf(outfile, FONT_STR);

#ifdef HORRIBLE_PMC_KLUDGE
        if (pmc_hack) {
            int x, x2, y;

            /* make a mono-header (instead of one header for each column) */
            x = SubpageOrigin[0][0];
            x2 = SubpageOrigin[SubpageMax-1][0] + ptswide;
            y = SubpageOrigin[0][1] + 30 +  /* was HeaderOffset: 12*(2 or 3) */
                 YINC*LineMax - 10 - 6;     /* extra 1/2 line */
            fprintf(outfile, "save\n");
            fprintf(outfile, "newpath %d %d moveto %d %d lineto %d %d lineto "
              "%d %d lineto eofill\n", x-1, y,  x2+13, y,  x2+13, y+YINC+4,
              x-1, y+YINC+4);
            fprintf(outfile, "1 1 1 setrgbcolor\n");
            fprintf(outfile, "/Times-Bold findfont 12 scalefont setfont\n");
            fprintf(outfile, "%d %d moveto (PMC) show\n", x+6, y+5);
            fprintf(outfile, "%d %d moveto (cool, huh?) centershow\n",
              (x+6 + x2+6) >> 1, y+5);
            fprintf(outfile, "%d %d moveto (%s) rightshow\n", x2+6, y+5, when);
            fprintf(outfile, "restore\n");
        }
#endif /* HORRIBLE_PMC_KLUDGE */

    /*-----------------------------------------------------------------------
        Loop over logical pages (sub-pages).
      -----------------------------------------------------------------------*/

        while (!end_physical_page) {
            int x, y, eol, discard, blank_line, blank_page;
            int line_counter, end_logical_page;

            // [^.%A-Za-z0-9_]x[^.%A-Za-z0-9_]  [^.%A-Za-z0-9_]y[^.%A-Za-z0-9_]
            // [^.%A-Za-z0-9_]c[^.%A-Za-z0-9_]

            ++subpagenum;

        /*-------------------------------------------------------------------
            Set up and print the header line.
          -------------------------------------------------------------------*/

            if (header) {
                char *p, pagehdr[16];
                int n, n1, pagenum_len, num_esc_chars, include_date=TRUE;
                float gr;

                x = SubpageOrigin[subpage_counter][0];
                y = SubpageOrigin[subpage_counter][1] + YINC*LineMax + 30 - 10;
                                   /* 30 was YINC*HeaderOffset:  12*(2 or 3) */

                /* shade header line */
                switch (gray) {
                    case 1:   /* empty box only */
                        fprintf(outfile, "0 0 0 setrgbcolor\nnewpath %d %d "
                          "moveto %d %d lineto %d %d lineto %d %d lineto "
                          "closepath stroke\n",
                          x, y,
                          x+ptswide+12, y,
                          x+ptswide+12, y+YINC+4,
                          x, y+YINC+4);
                        break;
                    case 2:
                    case 3:
                        gr = (float)(99 - gray) * .01;           /* .97, .96 */
                        fprintf(outfile, "%.2f %.2f %.2f", gr, gr, gr);
                        break;
                    case 4:
                    case 5:
                        gr = (float)(145 - 15*gray) * .01;       /* .85, .70 */
                        fprintf(outfile, "%.2f %.2f %.2f", gr, gr, gr);
                        break;
                    case 6:
                    case 7:
                    case 8:
                    case 9:
                        gr = (float)(9 - gray) * .15;    /* .45, .30, .15, 0 */
                        fprintf(outfile, "%.2f %.2f %.2f", gr, gr, gr);
                }
                /* make gray boxes slightly wider than (hollow) border boxes */
                if (gray > 1)
                    fprintf(outfile, " setrgbcolor\nnewpath %d %d moveto "
                      "%d %d lineto %d %d lineto %d %d lineto eofill\n",
                      x-1, y,
                      x+ptswide+13, y,
                      x+ptswide+13, y+YINC+4,
                      x-1, y+YINC+4);
                if (gray < 6)
                    fprintf(outfile, "0 0 0 setrgbcolor\n");   /* black text */
                else
                    fprintf(outfile, "1 1 1 setrgbcolor\n");   /* white text */

                /* now prepare the header text */

                sprintf(pagehdr, "Page %d", subpagenum);
                pagenum_len = strlen(pagehdr);

                n1 = strlen(fullname);
                if ((CharMax - (n1 + pagenum_len)) < 2) {
                    int num_to_copy;

                    /* filename (n1) is too long; want to copy only last
                     * portion and add "..." at beginning */
                    num_to_copy = CharMax - pagenum_len - 2 /* spaces */
                                                        - 3 /* dots */ ;
                    p = fullname + n1 - num_to_copy;
                    outbuf[0] = '.';
                    outbuf[1] = '.';
                    outbuf[2] = '.';
                    n = PS_strncpy((char *)outbuf+3, p, BUFSIZE-3) + 3;
                    n1 = num_to_copy + 3;
                } else {
                    /* cp fullname to outbuf, escaping any '(', ')', or '\\' */
                    n = PS_strncpy((char *)outbuf, fullname, BUFSIZE);
                }
                num_esc_chars = n - n1;

                outptr = outbuf + n;		/* points at '\0' */
                /* note that outptr is no longer valid for counting displayed
                 * output chars due to possible escape characters */

                if ((n = CharMax - (n1 + DATE_LEN + pagenum_len)) < 4) {
                    /* header line doesn't have sufficient room for
                     *  fullname + 2 spaces + date + 2 spaces + "Page #" */
                    include_date = FALSE;
                    /* definitely no room for date, but because of num_to_copy
                     * section above, we know we do have room for "Page ##" */
                    n = (CharMax - (n1 + pagenum_len) + 1) / 2;  /* round up */
                    p = (char *)outptr + n;     /* ...so no empty char-slots */
                } else {
                   /* have sufficient room in which to fit date string */
                    n = (n + 1) / 2;
                    p = (char *)outptr + n + DATE_LEN;
                }
                /* [n is no longer the number of displayed fullname chars] */
                while (n-- > 0) {
                    *outptr++ = ' ';	/* pad with spaces... */
                    *p++ = ' ';		/* ...on both sides of date string */
                }
                if (include_date) {
                    for (p = when;  *p;  ++p)	/* copy date string */
                        *outptr++ = *p;
                }
                outptr = outbuf + CharMax - pagenum_len + num_esc_chars;
                p = pagehdr;			/* copy page number... */
                while ((*outptr++ = *p++)) ;	/*  including trailing \0 */

                fprintf(outfile, SQUISH show_string, x+6, y+5, outbuf);
                if (gray > 5)
                    fprintf(outfile, "0 0 0 setrgbcolor\n");   /* black text */
            }

        /*-------------------------------------------------------------------
            Draw the border box.
          -------------------------------------------------------------------*/

            if (border) {
                int ptshigh;

                x = SubpageOrigin[subpage_counter][0];
                y = SubpageOrigin[subpage_counter][1];
                ptshigh = YINC*LineMax + 8;

                fprintf(outfile, "newpath %d %d moveto %d %d lineto %d %d "
                  "lineto %d %d lineto closepath stroke\n",
                  x, y,
                  x+ptswide+12, y,
                  x+ptswide+12, y+ptshigh,
                  x, y+ptshigh);
            }

        /*-------------------------------------------------------------------
            Loop over lines on current logical (sub)page.
          -------------------------------------------------------------------*/

            x = SubpageOrigin[subpage_counter][0] + 6;
            y = SubpageOrigin[subpage_counter][1] + YINC*LineMax - 5;

            line_counter = 0;
            end_logical_page = FALSE;
            blank_page = TRUE;

            while (!end_logical_page) {
                int c=0, char_counter;

                /* set up for next line */
                eol = discard = FALSE;
                if (!wrapped)
                    ++lineno;   /* used for -n option, not for page length */
                if (numbers) {
#if 0
                    if (wrapped)
                        sprintf((char *)outbuf, "       ");
                    else
                        sprintf((char *)outbuf, "%5d  ", lineno);
#endif
                    sprintf((char *)outbuf, wrapped? "       " : "%5d  ",
                      lineno);
                    char_counter = NUMBERS_WIDTH;
                    outptr = outbuf + NUMBERS_WIDTH;
                    blank_line = blank_page = FALSE;
                } else {
                    char_counter = 0;
                    outptr = outbuf;
                    blank_line = TRUE;
                }

            /*---------------------------------------------------------------
                Get characters from the input file until an end-of-line char-
                acter (line feed, form feed) is found or until the end of the
                file is reached.  Check special cases in order of expected
                frequency.
              ---------------------------------------------------------------*/

#ifdef VMS
                while (!eol  &&  (c = (*get_c)(infile)) != EOF)
#else
                while (!eol  &&  (c = getc(infile)) != EOF)
#endif
                {
                    switch (c) {
                        case LF:
                          wrapped = FALSE;
                          eol = TRUE;
                          break;
                        case TAB:
                          if (!discard) {
                              if ((outptr + 
                                  (tabsize - (char_counter-num_offset+7)%tabsize
                                   - 1)) >= outbuf + BUFSIZE)
                              {
                                  eol = TRUE;  /* else buffer overrun */
                                  DebugPrint((stderr, DEBUG_WARNING_OVERRUN));
                                  break;
                              }
                              do {
                                  *outptr++ = ' ';
                              } while ((++char_counter - num_offset) % tabsize);
                              /* a few extras at end of line won't matter */
                              if (char_counter >= CharMax) {
                                  if (wrap) {
                                      eol = TRUE;
                                      wrapped = TRUE;
                                  } else
                                      discard = TRUE;
                              }
                          }
                          break;
                        case BS:
                          --outptr;
                          if (outptr < outbuf) {
                              outptr = outbuf;   /* else buffer underrun */
                              DebugPrint((stderr, DEBUG_WARNING_UNDERRUN));
                          }
                          --char_counter;
                          if (char_counter < 0)
                              char_counter = 0;
                          break;
#if 0
                        case SOH:	/* ^A (01h): used in printfile format */
                          if ((c = getc(infile)) == VMS_RI)
                              eol = TRUE;
                          else
                              ungetc(c, infile);
                          break;
#endif
                        case FF:
                          wrapped = FALSE;
                          // BUG?  blank_page is true if logical subpage has
                          //   no printable characters on it yet...
                          //   This appears to "reuse" such a page if a form-
                          //   feed is encountered => not entirely correct, but
                          //   probably OK to keep (until someone complains...)
                          if (blank_page) {
                              y = SubpageOrigin[subpage_counter][1] +
                                    YINC*LineMax - 5;
                              line_counter = 0;
                          } else
                              end_logical_page = eol = TRUE;
                          if ((outptr-outbuf) == num_offset) {
                              blank_line = TRUE;	/* don't number FF */
                              --lineno;
                          }
                          break;
                    /*  case CR:
                        case SOH:
                        case VMS_RI:	skip these dudes (in default section)
                        case DOS_EOF:
                        case NUL:
                          break;
                     */
                        case '(':	/* escape special PS chars and fall */
                        case ')':	/*  through, but don't count them */
                        case '\\':
                          if (!discard)
                              *outptr++ = '\\';		/* fall through */
                              /* no need for buffer-overrun check here since
                               * _previous_ iteration's check (default, below)
                               * ensures sufficient space */
                        default:
                          if (!discard && (c > 31) && (c < 127)) {
                              *outptr++ = c;
                              if (c != ' ')
                                  blank_line = blank_page = FALSE;
                              /* this test is a bit too strict to accommodate a
                               * '\\' + escaped character + '\0' in the *next*
                               * iteration of the loop: */
                              if (outptr > outbuf + BUFSIZE - 3) {
                                  eol = TRUE;  /* else buffer overrun */
                                  DebugPrint((stderr, DEBUG_WARNING_OVERRUN));
                              }
                              /* even with buffer-overrun truncation, can still
                               * check for CRs, etc. */
                              if (++char_counter == CharMax) {
                                  if (wrap) {
                                      eol = TRUE;
                                      if (((c = getc(infile)) == EOF) ||
                                          (c == CR) || (c == LF) || (c == FF))
                                      {
                                          wrapped = FALSE; /* else extra line */
                                          if (c == CR) {   /* CR/LF check */
                                              c = getc(infile);
                                              /* eat LFs only; others go back */
                                              if (c != LF && c != EOF)
                                                  ungetc(c, infile);
                                          } else if (c == FF)
                                              ungetc(c, infile);
                                      } else {
                                          wrapped = TRUE;
                                          ungetc(c, infile);
                                      }
                                  } else
                                      discard = TRUE;
                              }
                          }
                          break;
                    } /* end switch */
                } /* end while (c != EOF  &&  !eol) */

                /* print line, if any non-white characters to print; if line
                 * is nothing but line number (== num_offset), print only if
                 * actually read an end-of-line char (i.e., not at EOF) */
                if (!blank_line && (eol || (outptr-outbuf > num_offset))) {
                    *outptr = '\0';
                    fprintf(outfile, SQUISH show_string, x, y, outbuf);
                }
                y -= YINC;

                if (c == EOF) {
                    /* done with file:  eject page and leave */
                    fprintf(outfile,
                      "showpage restore restore\n%%%%Trailer\n%%%%Pages: %d\n",
                      phys_pagenum);
                    return;
                }

                if (++line_counter == LineMax)
                    end_logical_page = TRUE;

            } /* end while (!end_logical_page) */

            if (++subpage_counter == SubpageMax) {
                int ch;

                fprintf(outfile, "showpage restore\n");
                end_physical_page = TRUE;
                /* check for next char; if no more, don't print blank page */
#ifdef DEBUG
                /* quit after one physical sheet of paper is full */
                if (debug > 1 && ((doublesided && phys_pagenum == 2) ||
                                  (!doublesided && phys_pagenum == 1)))
                {
                    fprintf(outfile, "restore\n%%%%Trailer\n%%%%Pages: %d\n",
                      phys_pagenum);
                    return;
                } else
#endif
                if ((ch = getc(infile)) == EOF) {
                    fprintf(outfile, "restore\n%%%%Trailer\n%%%%Pages: %d\n",
                      phys_pagenum);
                    return;
                } else
                    ungetc(ch, infile);
            }
        } /* end while (!end_physical_page) */
    } /* end for-loop (pages) */
} /* end function print_file() */





/***************************/
/*  Function PS_strncpy()  */  /* much like strncpy() except escapes PS chars */
/***************************/  /* and returns count of chars, not dest ptr */

int PS_strncpy(char *dest, const char *src, int n)
{
    const char *p = src - 1;
    char *q = dest;
    int i = n - 1;   /* leave room for trailing null */

    if (n < 0) {
        fprintf(stderr, "gpr error:  internal logic error (n < 0)\n");
        exit(1);
    } else if (n == 0) {
        return 0;
    }

    /* Unix allows filenames with any or all of '(', ')', and '\\' in them,
     * while MinGW version of getcwd() returns paths with '\\' rather than
     * '/'.  (Unix and Cygwin versions both use '/'.)  So copy source string
     * but escape any special PostScript characters while doing so.  (Also
     * check for buffer overruns, since string could double in size.) */

    while (*++p && i > 0) {
        if (*p == '(' || *p == ')' || *p == '\\') {
            if (i >= 2) {   /* must have room for both '\\' and escaped char */
                *q++ = '\\';
                --i;
            } else
                break;
        }
        *q++ = *p;
        --i;
    }
    *q = '\0';

    return (q - dest);
}





#ifdef VMS

/************************/
/*  Function handler()  */
/************************/

void handler(void)
{
    fprintf(stderr, "Interrupt:  restoring file attributes\n");
    change_file_attributes(filename, RESTORE);
    exit(0);
}





/*************************/
/*  Function vms_getc()  */
/*************************/

int vms_getc(FILE *fp)
{
    static int reading=FALSE, char_count, odd;

/*---------------------------------------------------------------------------
    If not in the middle of a line, must be pointing to the character count
    for the next line, so get that first.  If count is zero, return a line
    feed; else return the first character in the line.  But before doing
    that, make sure any NUL padding has been removed.
  ---------------------------------------------------------------------------*/

    if (!reading) {
        int c;

        if ((char_count = getc(infile)) == EOF)
            return char_count;
        c = getc(infile);	/* get rid of padding (char_count += c<<8 ?) */

        if (char_count == 0)
            return LF;
        if ((char_count == 255) && (c == 255))
            return EOF;

        /* all other cases:  starting a line */
        reading = TRUE;
        odd = (char_count % 2)?  TRUE : FALSE;
        return getc(infile);
    }

/*---------------------------------------------------------------------------
    While still characters left in line, get them and return to calling func-
    tion.  Otherwise reset reading flag, remove NUL padding (if any), and
    return a newline.
  ---------------------------------------------------------------------------*/

    else {		/* still counting down on current line */
        if (--char_count)
            return getc(infile);
        else {
            reading = FALSE;
            if (odd)
                getc(infile);	/* usually zero, but NOT ALWAYS! */
            return LF;
        }
    }

} /* end function vms_getc() */





#include "VMS_attr.c"

#endif /* VMS */
